#include "client_app.h"  // NOLINT(build/include)

#include <string>

#include "include/cef_cookie.h"
#include "include/cef_process_message.h"
#include "include/cef_task.h"
#include "include/cef_v8.h"
#include "util.h"  // NOLINT(build/include)
#include "cefclient/client_transfer.h"
#include "cefclient/client_app_js.h"

// #include <QDebug>

namespace
{

// Handles the native implementation for the client_app extension.
class ClientAppExtensionHandler : public CefV8Handler
{
public:
    explicit ClientAppExtensionHandler( CefRefPtr<ClientApp> client_app )
       : client_app_( client_app )
    {
    }

    virtual bool Execute( const CefString& name,
                          CefRefPtr<CefV8Value> object,
                          const CefV8ValueList& arguments,
                          CefRefPtr<CefV8Value>& retval,
                          CefString& exception )
    {
        bool handled = false;

        if ( name == JS_FUNC_SENDMESSAGE && arguments.size() >= 1 )
        {
            CefRefPtr<CefBrowser> browser =
               CefV8Context::GetCurrentContext()->GetBrowser();
            ASSERT( browser.get() );

            CefRefPtr<CefProcessMessage> message = CefProcessMessage::Create( name );

            cefclient::SetList( arguments, message->GetArgumentList() );

            browser->SendProcessMessage( PID_BROWSER, message );
            handled = true;
        }

        if ( !handled ) exception = "Invalid method arguments";

        return true;
    }

private:
    CefRefPtr<ClientApp> client_app_;

    IMPLEMENT_REFCOUNTING( ClientAppExtensionHandler );
};

}  // namespace


ClientApp::ClientApp()
{
    CreateBrowserDelegates( browser_delegates_ );
    CreateRenderDelegates( render_delegates_ );

    // Default schemes that support cookies.
    //cookieable_schemes_.push_back("http");
    //cookieable_schemes_.push_back("https");
}

void ClientApp::OnContextInitialized()
{
    // Register cookieable schemes with the global cookie manager.
    //CefRefPtr<CefCookieManager> manager = CefCookieManager::GetGlobalManager();
    //ASSERT(manager.get());
    //manager->SetSupportedSchemes(cookieable_schemes_);

    BrowserDelegateSet::iterator it = browser_delegates_.begin();
    for (; it != browser_delegates_.end(); ++it ) (*it)->OnContextInitialized( this );
}

void ClientApp::OnBeforeChildProcessLaunch(
   CefRefPtr<CefCommandLine> command_line )
{
    BrowserDelegateSet::iterator it = browser_delegates_.begin();
    for (; it != browser_delegates_.end(); ++it ) (*it)->OnBeforeChildProcessLaunch( this, command_line );
}

void ClientApp::OnRenderProcessThreadCreated(
   CefRefPtr<CefListValue> extra_info )
{
    BrowserDelegateSet::iterator it = browser_delegates_.begin();
    for (; it != browser_delegates_.end(); ++it ) (*it)->OnRenderProcessThreadCreated( this, extra_info );
}

void ClientApp::OnRenderThreadCreated( CefRefPtr<CefListValue> extra_info )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnRenderThreadCreated( this, extra_info );
}

void ClientApp::OnWebKitInitialized()
{
//// Register the client_app extension.
//  std::string app_code =
//  "var app;"
//  "if (!app) {"
//  "  app = {"
//  "    sendMessage: function(/*one or more*/) {"
//  "      native function sendMessage();"
//  "      return sendMessage.apply(this, Array.prototype.slice.call(arguments));"
//  "    }"
//  "  }"
//  "}";
//CefRegisterExtension("v8/app", app_code,
//    new ClientAppExtensionHandler(this));

    // Execute delegate callbacks.
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnWebKitInitialized( this );
}

void ClientApp::OnBrowserCreated( CefRefPtr<CefBrowser> browser )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnBrowserCreated( this, browser );
}

void ClientApp::OnBrowserDestroyed( CefRefPtr<CefBrowser> browser )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnBrowserDestroyed( this, browser );
}

CefRefPtr<CefLoadHandler> ClientApp::GetLoadHandler()
{
    CefRefPtr<CefLoadHandler> load_handler;

    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end() && !load_handler.get(); ++it ) load_handler = (*it)->GetLoadHandler( this );

    return load_handler;
}

bool ClientApp::OnBeforeNavigation( CefRefPtr<CefBrowser> browser,
                                    CefRefPtr<CefFrame> frame,
                                    CefRefPtr<CefRequest> request,
                                    NavigationType navigation_type,
                                    bool is_redirect )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it )
    {
        if ( (*it)->OnBeforeNavigation( this, browser, frame, request,
                                        navigation_type, is_redirect ) )
        {
            return true;
        }
    }

    return false;  // Allow the navigation to proceed.
}

void ClientApp::OnContextCreated( CefRefPtr<CefBrowser> browser,
                                  CefRefPtr<CefFrame> frame,
                                  CefRefPtr<CefV8Context> context )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnContextCreated( this, browser, frame, context );
}

void ClientApp::OnContextReleased( CefRefPtr<CefBrowser> browser,
                                   CefRefPtr<CefFrame> frame,
                                   CefRefPtr<CefV8Context> context )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnContextReleased( this, browser, frame, context );
}

void ClientApp::OnUncaughtException( CefRefPtr<CefBrowser> browser,
                                     CefRefPtr<CefFrame> frame,
                                     CefRefPtr<CefV8Context> context,
                                     CefRefPtr<CefV8Exception> exception,
                                     CefRefPtr<CefV8StackTrace> stackTrace )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it )
    {
        (*it)->OnUncaughtException( this, browser, frame, context, exception,
                                    stackTrace );
    }
}

void ClientApp::OnFocusedNodeChanged( CefRefPtr<CefBrowser> browser,
                                      CefRefPtr<CefFrame> frame,
                                      CefRefPtr<CefDOMNode> node )
{
    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end(); ++it ) (*it)->OnFocusedNodeChanged( this, browser, frame, node );
}

bool ClientApp::OnProcessMessageReceived(
   CefRefPtr<CefBrowser> browser,
   CefProcessId source_process,
   CefRefPtr<CefProcessMessage> message )
{
    ASSERT( source_process == PID_BROWSER );

    bool handled = false;

    RenderDelegateSet::iterator it = render_delegates_.begin();
    for (; it != render_delegates_.end() && !handled; ++it )
    {
        handled = (*it)->OnProcessMessageReceived( this, browser, source_process, message );
    }


//  qDebug() << __FILE__ << __FUNCTION__ << __LINE__ << QString( CefString( message->GetName() ).ToString().c_str() ) << "handled" << handled;

    return handled;
}